from ..registry import register_model  # noqa
from ..builders import build_model  # noqa
from .model import Model, ModelConfig  # noqa
from .model_outputs import *
from .backbone import *
from .image2text import *
from .mask_filling import *
from .sequence_labeling import *
from .speech_recognition import *
from .text_classification import *
from .text_detection import *
from .text_generation import *



from __future__ import annotations

import os
import tempfile
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Tuple

import numpy as np
import torch
from huggingface_hub import create_repo, hf_hub_download, upload_file

from ...builders import build_preprocessor
from ...configs import PreprocessorConfig
from ...constants import (
    DEFAULT_TOKENIZER_CONFIG_FILE,
    DEFAULT_TOKENIZER_FILE,
    HEZAR_CACHE_DIR,
    Backends,
    PaddingType,
)
from ...utils import (
    Logger,
    convert_batch_dict_dtype,
    is_backend_available,
    pad_batch_items,
)
from ..preprocessor import Preprocessor


if is_backend_available(Backends.TOKENIZERS):
    from tokenizers import Tokenizer as HFTokenizer
    from tokenizers.decoders import Decoder
    from tokenizers.models import Model

logger = Logger(__name__)


@dataclass
class TokenizerConfig(PreprocessorConfig):
    """
    Configuration for the Tokenizer.

    Args:
        truncation_side (str): Truncation direction for tokenization.
        stride (int): Stride for tokenization.
        padding_side (str): Padding direction for tokenization.
        pad_to_multiple_of (int): Pad to a multiple of this value.
        pad_token_type_id (int): ID of the padding token type.
        bos_token (str): Beginning of sequence token.
        eos_token (str): End of sequence token.
        unk_token (str): Unknown token.
        sep_token (str): Separator token.
        pad_token (str): Padding token.
        cls_token (str): Classification token.
        mask_token (str): Mask token.
        additional_special_tokens (List[str]): Additional special tokens.
    """

    name = "tokenizer"
    max_length: int = "deprecated"
    truncation: str = "deprecated"
    truncation_side: str = None
    padding: str = "deprecated"
    padding_side: str = None
    stride: int = None
    pad_to_multiple_of: int = "deprecated"
    pad_token_type_id: int = 0
    bos_token: str = None
    eos_token: str = None
    unk_token: str = None
    sep_token: str = None
    pad_token: str = None
    cls_token: str = None
    mask_token: str = None
    additional_special_tokens: List[str] = None

    def __post_init__(self):
        super().__post_init__()
        if self.max_length != "deprecated":
            logger.warning(
                "Setting `max_length` in the tokenizer config is deprecated and will be removed in the future!"
            )
        if self.padding != "deprecated":
            logger.warning("Setting `padding` in the tokenizer config is deprecated and will be removed in the future!")
        if self.truncation != "deprecated":
            logger.warning(
                "Setting `truncation` in the tokenizer config is deprecated and will be removed in the future!"
            )


class Tokenizer(Preprocessor):
    """
    Base tokenizer class. Mostly copied from :class:`~tokenizers.implementations.BaseTokenizer`.

    Args:
        config: A TokenizerConfig instance.
        tokenizer_file (str): A tokenizer.json file to load the whole tokenizer from.
        **kwargs: Extra config parameters that merge into the main config.
    """

    required_backends: List[str | Backends] = []

    tokenizer_filename = DEFAULT_TOKENIZER_FILE
    tokenizer_config_filename = DEFAULT_TOKENIZER_CONFIG_FILE
    token_ids_name = "token_ids"
    uncastable_keys = ["word_ids", "tokens", "offsets_mapping"]

    def __init__(self, config: TokenizerConfig, tokenizer_file=None, **kwargs):
        super().__init__(config, **kwargs)
        self._tokenizer = self.from_file(tokenizer_file) if tokenizer_file is not None else self.build()
        self.special_tokens = self._get_all_special_tokens()

    def _get_all_special_tokens(self):
        """
        Get a list of all special tokens.

        Returns:
            List[str]: List of special tokens.
        """
        _special_tokens = [
            self.config.bos_token,
            self.config.eos_token,
            self.config.unk_token,
            self.config.sep_token,
            self.config.pad_token,
            self.config.cls_token,
            self.config.mask_token,
        ]
        _special_tokens = [token for token in _special_tokens if token in self.vocab]

        if self.config.additional_special_tokens is not None:
            for token in self.config.additional_special_tokens:
                if token not in _special_tokens:
                    _special_tokens.append(token)

        valid_tokens = [token for token in _special_tokens if token is not None]
        return valid_tokens

    @staticmethod
    def from_file(path):
        """
        Create a tokenizer from a file.

        Args:
            path (str): Path to the tokenizer file.

        Returns:
            HFTokenizer: The created tokenizer.
        """
        tokenizer = HFTokenizer.from_file(path)
        return tokenizer

    def build(self):
        """
        Build the tokenizer.

        Returns:
            HFTokenizer: The built tokenizer.
        """
        raise NotImplementedError

    def encode(self, inputs, is_pretokenized: bool = False, add_special_tokens: bool = True, **kwargs):
        """
        Tokenize a list of inputs (could be raw or tokenized inputs).

        Args:
            inputs: List of inputs.
            is_pretokenized: Whether the inputs are already tokenized.
            add_special_tokens: Whether to add special tokens to the inputs. Defaults to True.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Dict]: List of dictionaries containing tokenized inputs.
        """
        if isinstance(inputs, str):
            inputs = [inputs]
        elif isinstance(inputs, list) and is_pretokenized:
            if isinstance(inputs[0], str):
                inputs = [inputs]
        return self._tokenizer.encode_batch(inputs, is_pretokenized, add_special_tokens)

    def decode(self, ids: List[int], skip_special_tokens: bool = True, **kwargs):
        """
        Decode a list of token IDs.

        Args:
            ids (List[int]): List of token IDs.
            skip_special_tokens (bool): Whether to skip special tokens during decoding.
            **kwargs: Additional keyword arguments.

        Returns:
            List[str]: List of decoded strings.
        """
        if isinstance(ids[0], int):
            ids = [ids]
        if isinstance(ids, (torch.Tensor, np.ndarray)):
            ids = ids.tolist()
        return self._tokenizer.decode_batch(ids, skip_special_tokens=skip_special_tokens)

    def pad_encoded_batch(
        self,
        inputs,
        padding: str | PaddingType = None,
        max_length: Optional[int] = None,
        truncation: bool = True,
        return_tensors: Optional[str] = None,
        include_keys: Optional[List[str]] = None,
        exclude_keys: List = None,
    ):
        """
        Pad a batch of encoded inputs.

        Args:
            inputs: Input batch of encoded tokens.
            padding (str | PaddingType): Padding type.
            max_length (Optional[int]): Max input length (only if padding is set to "max_length").
            truncation (bool): Whether to allow truncation.
            return_tensors (Optional[str]): The type of tensors to return.
            include_keys: (Optional[List[str]]): Only pad these given set of keys
            exclude_keys (List): A list of keys to exclude when padding.

        Returns:
            Dict: Padded inputs.
        """
        if isinstance(inputs, (list, tuple)) and isinstance(inputs[0], Mapping):
            inputs = {key: [example[key] for example in inputs] for key in inputs[0].keys()}

        exclude_keys = exclude_keys or []
        exclude_keys += self.uncastable_keys  # avoid possible errors
        inputs = convert_batch_dict_dtype(inputs, dtype="list", skip_keys=exclude_keys)

        include_keys = include_keys or list(inputs.keys())

        for key, batch in inputs.items():
            if key in exclude_keys:
                continue
            if key in include_keys:
                pad_id = 0 if key == "attention_mask" else self.pad_token_id
                padded_ids = pad_batch_items(
                    inputs[key],
                    padding=padding,
                    padding_side=self.config.padding_side,
                    pad_id=pad_id,
                    max_length=max_length,
                    truncation=truncation,
                )
                inputs[key] = padded_ids

        inputs = convert_batch_dict_dtype(inputs, dtype=return_tensors, skip_keys=exclude_keys)

        return inputs

    def __call__(
        self,
        inputs: List[str] | List[Tuple[str, str]],
        device: str | torch.device = None,
        add_special_tokens: bool = True,
        padding=None,
        truncation=None,
        max_length: int = None,
        return_tensors: str = "list",
        stride: int = 0,
        is_split_into_words: bool = False,
        pad_to_multiple_of: int = None,
        return_tokens: bool = None,
        return_token_type_ids: bool = None,
        return_attention_mask: bool = True,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        return_word_ids: bool = False,
        verbose: bool = True,
        **kwargs,
    ):
        """
        Tokenize a batch of string inputs and return the relevant properties e.g, token ids, attention mask, etc.

        Args:
            inputs: A list of string inputs to tokenize
            add_special_tokens: Whether to add special tokens or not
            padding: Determines how to pad inputs
            truncation: Determines how to truncate inputs
            max_length: Max input length of the sequences
            return_tensors: The type of the returning tensors in the batch e.g, pt, np, list
            stride: Stride level
            is_split_into_words: Are inputs pre-tokenized or raw string inputs
            pad_to_multiple_of: Pad inputs by a factor of this value
            return_tokens: Whether to return tokens lists
            return_token_type_ids: Whether to return token type ids
            return_attention_mask: Whether to return attention masks
            return_overflowing_tokens: Whether to return overflowing tokens
            return_special_tokens_mask: Whether to return special tokens mask
            return_offsets_mapping: Whether to return offsets
            return_length: Whether to return input lengths
            **kwargs: Extra arguments reside here and therefore ignored

        Returns:
            A dictionary of encoded inputs like
                {"token_ids": [batch_size x input_len], "attention_mask": [batch_size x input_len], ...}
        """
        if isinstance(inputs, list) and not len(inputs):
            raise ValueError("Tokenizer cannot process an empty list!")

        if "padding_strategy" in kwargs:
            logger.warning(
                "`padding_strategy` was deprecated in favor of `padding`!"
                " This warning will change to an error in the future!"
            )
        if "truncation_strategy" in kwargs:
            logger.warning(
                "`truncation_strategy` was deprecated in favor of `truncation`!"
                " This warning will change to an error in the future!"
            )

        return_tensors = return_tensors or "list"

        # Convert to batch if input is a single string or a list of words (is split into words for sequence labeling)
        if isinstance(inputs, str) or (is_split_into_words and not isinstance(inputs[0], list)):
            inputs = [inputs]
            is_batch = False
        else:
            is_batch = True

        self.set_truncation_and_padding(
            padding=padding,
            truncation=truncation,
            padding_side=self.config.padding_side,
            truncation_side=self.config.truncation_side,
            max_length=max_length,
            stride=self.config.stride,
            pad_to_multiple_of=pad_to_multiple_of,
        )
        encodings = self.encode(
            inputs,
            add_special_tokens=add_special_tokens,
            is_pretokenized=is_split_into_words,
        )
        encodings_dict = [
            self._convert_encodings(
                encoding=encoding,
                original_text=inp if isinstance(inp, str) else None,
                return_tokens=return_tokens,
                return_token_type_ids=return_token_type_ids,
                return_attention_mask=return_attention_mask,
                return_overflowing_tokens=return_overflowing_tokens,
                return_special_tokens_mask=return_special_tokens_mask,
                return_offsets_mapping=return_offsets_mapping,
                return_length=return_length,
                return_word_ids=return_word_ids,
            )
            for encoding, inp in zip(encodings, inputs)
        ]
        # Permute output dict from [batch_0: Dict[key, value], ...] to Dict[key, [batch_0, batch_1, ...], ...]
        sanitized_outputs = {}
        for key in encodings_dict[0].keys():
            stack = [e for item in encodings_dict for e in item[key]]
            sanitized_outputs[key] = stack

        # If returning overflowing tokens, we need to return a mapping
        # from the batch idx to the original sample
        if return_overflowing_tokens:
            overflow_to_sample_mapping = []
            for i, encodings_ in enumerate(encodings_dict):
                overflow_to_sample_mapping += [i] * len(encodings_["input_ids"])
            sanitized_outputs["overflow_to_sample_mapping"] = overflow_to_sample_mapping

        # Squeeze tensor if the original input is a single string and return_tensors is `list`
        if (return_tensors == "list" or return_tensors is None) and not is_batch:
            sanitized_outputs = {
                key: value[0] if len(value) > 0 and isinstance(value[0], list) else value
                for key, value in sanitized_outputs.items()
            }

        outputs = convert_batch_dict_dtype(sanitized_outputs, dtype=return_tensors, skip_keys=self.uncastable_keys)

        if device and return_tensors == "torch":
            outputs = {k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in outputs.items()}

        return outputs

    def set_truncation_and_padding(
        self,
        padding=None,
        truncation=None,
        padding_side=None,
        truncation_side=None,
        max_length: int = None,
        stride: int = None,
        pad_to_multiple_of: int = None,
    ):
        # Set truncation and padding on the backend tokenizer
        if truncation == "no_truncation" or truncation is None or max_length is None:
            if self.truncation is not None:
                self.no_truncation()
        else:
            if truncation is True:
                truncation = "longest_first"
            target = {
                "max_length": max_length,
                "stride": stride,
                "strategy": truncation,
                "direction": truncation_side,
            }
            if self.truncation is None:
                current = None
            else:
                current = {k: self.truncation.get(k, None) for k in target}

            if current != target:
                self.enable_truncation(**target)

        if padding == "no_padding" or padding is None:
            if self.padding is not None:
                self.no_padding()
        else:
            target = {
                "length": max_length if padding == PaddingType.MAX_LENGTH else None,
                "direction": padding_side,
                "pad_id": self.token_to_id(self.pad_token),
                "pad_token": self.pad_token,
                "pad_type_id": self.config.pad_token_type_id,
                "pad_to_multiple_of": pad_to_multiple_of,
            }
            if self.padding != target:
                self.enable_padding(**target)

    def _convert_encodings(
        self,
        encoding,
        original_text: str | None = None,
        return_tokens: bool = None,
        return_token_type_ids: bool = None,
        return_attention_mask: bool = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        return_word_ids: bool = False,
    ):
        if return_overflowing_tokens and encoding.overflowing is not None:
            encodings = [encoding] + encoding.overflowing
        else:
            encodings = [encoding]

        encoding_dict = defaultdict(list)
        for e in encodings:
            encoding_dict[self.token_ids_name].append(e.ids)

            if return_token_type_ids:
                encoding_dict["token_type_ids"].append(e.type_ids)
            if return_attention_mask:
                encoding_dict["attention_mask"].append(e.attention_mask)
            if return_special_tokens_mask:
                encoding_dict["special_tokens_mask"].append(e.special_tokens_mask)
            if return_offsets_mapping:
                encoding_dict["offsets_mapping"].append(e.offsets)
            if return_length:
                encoding_dict["length"].append(len(e.ids))
            if return_tokens:
                tokens = self.get_tokens_from_encoding(e, original_text=original_text)
                encoding_dict["tokens"].append(tokens)
            if return_word_ids:
                encoding_dict["word_ids"].append(e.word_ids)

        return encoding_dict

    def convert_tokens_to_ids(self, tokens: str | List[str]) -> int | List[int]:
        if isinstance(tokens, str):
            tokens = [tokens]

        return [self._tokenizer.token_to_id(token) for token in tokens]

    def convert_ids_to_tokens(self, ids: int | List[int], skip_special_tokens: bool = False):
        if isinstance(ids, int):
            ids = [ids]
        tokens = []
        for index in ids:
            index = int(index)
            if skip_special_tokens and index in self.special_ids:
                continue
            tokens.append(self._tokenizer.id_to_token(index))
        return tokens

    def num_special_tokens_to_add(self, is_pair: bool) -> int:
        return self._tokenizer.num_special_tokens_to_add(is_pair)

    def get_vocab(self, with_added_tokens: bool = True) -> Dict[str, int]:
        return self._tokenizer.get_vocab(with_added_tokens=with_added_tokens)

    def get_vocab_size(self, with_added_tokens: bool = True) -> int:
        return self._tokenizer.get_vocab_size(with_added_tokens=with_added_tokens)

    def enable_padding(
        self,
        direction: str = "right",
        pad_to_multiple_of: int = None,
        pad_id: int = 0,
        pad_type_id: int = 0,
        pad_token: str = None,
        length: int = None,
    ):
        return self._tokenizer.enable_padding(
            direction=direction,
            pad_to_multiple_of=pad_to_multiple_of,
            pad_id=pad_id,
            pad_type_id=pad_type_id,
            pad_token=pad_token,
            length=length,
        )

    def no_padding(self):
        return self._tokenizer.no_padding()

    def enable_truncation(self, max_length, stride=0, strategy="longest_first", direction="right"):
        return self._tokenizer.enable_truncation(max_length, stride=stride, strategy=strategy, direction=direction)

    def no_truncation(self):
        return self._tokenizer.no_truncation()

    def add_tokens(self, tokens) -> int:
        return self._tokenizer.add_tokens(tokens)

    def add_special_tokens(self, special_tokens) -> int:
        return self._tokenizer.add_special_tokens(special_tokens)

    def token_to_id(self, token: str) -> int:
        return self._tokenizer.token_to_id(token)

    def id_to_token(self, id: int) -> str:
        return self._tokenizer.id_to_token(id)

    def get_added_vocab(self) -> Dict[str, int]:
        """
        Returns the added tokens in the vocabulary as a dictionary of token to index.

        Returns:
            `Dict[str, int]`: The added tokens.
        """
        base_vocab = self._tokenizer.get_vocab(with_added_tokens=False)
        full_vocab = self._tokenizer.get_vocab(with_added_tokens=True)
        added_vocab = {token: index for token, index in full_vocab.items() if token not in base_vocab}
        return added_vocab

    def __len__(self) -> int:
        """
        Size of the full vocabulary with the added tokens.
        """
        return self._tokenizer.get_vocab_size(with_added_tokens=True)

    def _is_byte_level(self) -> bool:
        """
        Return True if the underlying tokenizer uses a byte-level decoder (e.g. BPE).

        Byte-level tokenizers encode each byte of a UTF-8 character as a separate
        symbol, so ``encoding.tokens`` contains unreadable byte-level strings like
        ``'Ø§ÙĪ'`` instead of the original Persian word ``'او'``.  In that case we
        must reconstruct token strings by slicing the original input text with the
        offset mapping instead.
        """
        if self._tokenizer.decoder is None:
            return False
        return type(self._tokenizer.decoder).__name__ == "ByteLevel"

    def get_tokens_from_encoding(self, encoding, original_text: str | None = None):
        """
        Extract human-readable tokens from the HF ``Encoding`` object.

        Two strategies are used depending on the tokenizer type:

        * **Non-byte-level** (WordPiece, Unigram, …): ``encoding.tokens`` already
          holds the correct human-readable subword strings (e.g. ``'##زم'``), so we
          return them directly.
        * **Byte-level BPE**: ``encoding.tokens`` contains unreadable byte-encoded
          strings.  Here we slice ``original_text`` with ``encoding.offsets``, which
          always map back to the original input correctly.

        .. note::
            Never use ``tokenizer.decode(ids)`` + offsets here.  The decoder rewrites
            the text (removes zero-width non-joiners, re-spaces punctuation, …) so the
            character positions in the offsets no longer align.

        Args:
            encoding: An ``Encoding`` object returned by the HF tokenizers backend.
            original_text: The original input string, required for byte-level BPE.

        Returns:
            List[str]: A list of human-readable token strings.
        """
        if self._is_byte_level():
            if original_text is None:
                # Fallback: return raw tokens if original text is unavailable
                return list(encoding.tokens)
            return self.get_tokens_from_offsets(original_text, encoding.ids, encoding.offsets)
        return list(encoding.tokens)

    def get_tokens_from_offsets(
        self,
        text: str,
        ids: List[int],
        offsets_mapping: List[Tuple[int, int]],
    ):
        """
        Extract human-readable tokens by slicing *text* with the provided offsets.

        .. warning::
            *text* **must** be the same string that was originally fed to the tokenizer
            (or its normalised form if a normalizer is configured).  Never pass the
            output of ``tokenizer.decode(ids)`` here: the decoder rewrites the text
            (removes zero-width characters, re-spaces punctuation, …) so the character
            positions in *offsets_mapping* will no longer align correctly, resulting in
            garbled tokens.

        Args:
            text: The original (or normalised) input string.
            ids: Token ids corresponding to the offsets.
            offsets_mapping: A list of (start, end) character offset tuples into *text*.

        Returns:
            List[str]: A list of token strings.
        """
        if not isinstance(text, str):
            raise ValueError(f"Expected str type for `text`, got `{type(text)}({text})`")
        if isinstance(offsets_mapping, list) and not isinstance(offsets_mapping[0], Tuple):
            raise ValueError(f"Expected a list of tuples for `offsets_mapping`, got List[{type(offsets_mapping[0])}]")
        tokens = []
        for offset in offsets_mapping:
            offset_start, offset_end = offset
            tokens.append(text[offset_start:offset_end])
        for i, token in enumerate(tokens):
            if ids[i] in self.special_ids:
                tokens[i] = self._tokenizer.id_to_token(ids[i])
        return tokens

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        subfolder=None,
        config_filename=None,
        tokenizer_filename=None,
        cache_dir=None,
        **kwargs,
    ) -> "Tokenizer":
        """
        Load a tokenizer from a specified path or Hub repository.

        Args:
            cls: Class reference.
            hub_or_local_path: Path or Hub repository ID.
            subfolder: Subfolder containing tokenizer files.
            config_filename: Tokenizer config filename.
            tokenizer_filename: Tokenizer filename.
            cache_dir: Path to cache directory
            **kwargs: Additional arguments.

        Returns:
            Tokenizer: Loaded tokenizer.

        """
        tokenizer_filename = tokenizer_filename or cls.tokenizer_filename
        config_filename = config_filename or cls.tokenizer_config_filename
        subfolder = subfolder or cls.preprocessor_subfolder
        cache_dir = cache_dir or HEZAR_CACHE_DIR

        config = TokenizerConfig.load(
            hub_or_local_path,
            filename=config_filename,
            subfolder=subfolder,
            cache_dir=cache_dir,
        )

        if os.path.isdir(hub_or_local_path):
            tokenizer_path = os.path.join(hub_or_local_path, subfolder, tokenizer_filename)
        else:
            tokenizer_path = hf_hub_download(
                hub_or_local_path,
                filename=tokenizer_filename,
                subfolder=subfolder,
                cache_dir=cache_dir,
            )
        tokenizer = build_preprocessor(config.name, config, tokenizer_file=tokenizer_path, **kwargs)
        return tokenizer

    def save(self, path, save_config=True, pretty=True):
        """
        Save the tokenizer and its configuration.

        Args:
            path (str): Path to save the tokenizer.
            save_config (bool): Whether to save the configuration.
            pretty (bool): Whether to format the saved JSON file with indentation.

        """
        os.makedirs(path, exist_ok=True)
        # save config
        if save_config:
            self.config.vocab_size = self.get_vocab_size(with_added_tokens=True)
            self.config.save(path, filename=self.tokenizer_config_filename, subfolder=self.preprocessor_subfolder)
        # save tokenizer.json
        save_path = os.path.join(path, self.preprocessor_subfolder, self.tokenizer_filename)
        self._tokenizer.save(save_path, pretty=pretty)

    def push_to_hub(
        self,
        repo_id,
        commit_message=None,
        subfolder=None,
        tokenizer_filename=None,
        config_filename=None,
        private=False,
    ):
        """
        Push tokenizer and config to the Hub

        Args:
            repo_id: The path (id or repo name) on the hub
            commit_message: Commit message for this push
            subfolder: subfolder to save the files
            tokenizer_filename: tokenizer filename
            config_filename: tokenizer config filename
            private: If the repo should be private (ignored if the repo exists)
        """
        subfolder = subfolder or self.preprocessor_subfolder
        tokenizer_filename = tokenizer_filename or self.tokenizer_filename
        config_filename = config_filename or self.tokenizer_config_filename

        # create remote repo
        create_repo(repo_id, exist_ok=True, private=private)
        # save to tmp and prepare for push
        cache_path = tempfile.mkdtemp()
        # save tokenizer.json
        tokenizer_save_path = os.path.join(cache_path, subfolder, tokenizer_filename)
        self.save(cache_path, pretty=True)

        if commit_message is None:
            commit_message = "Hezar: Upload tokenizer and config"

        # upload config
        self.config.push_to_hub(
            repo_id=repo_id,
            filename=config_filename,
            subfolder=subfolder,
            commit_message=commit_message,
        )
        # upload tokenizer
        upload_file(
            repo_id=repo_id,
            path_or_fileobj=tokenizer_save_path,
            repo_type="model",
            path_in_repo=f"{subfolder}/{tokenizer_filename}",
            commit_message=commit_message,
        )
        logger.log_upload_success(
            name=f"{self.__class__.__name__}(name={self.config.name})",
            target_path=os.path.join(repo_id, subfolder, tokenizer_filename),
        )

    @property
    def model(self) -> "Model":
        return self._tokenizer.model

    @model.setter
    def model(self, model: "Model"):
        self._tokenizer.model = model  # noqa

    @property
    def decoder(self) -> "Decoder":
        return self._tokenizer.decoder

    @decoder.setter
    def decoder(self, decoder: "Decoder"):
        self._tokenizer.decoder = decoder  # noqa

    @property
    def padding(self):
        return self._tokenizer.padding

    @property
    def truncation(self) -> dict:
        return self._tokenizer.truncation

    @property
    def vocab(self):
        return self._tokenizer.get_vocab(with_added_tokens=True)

    @property
    def vocab_size(self) -> int:
        """
        `int`: Size of the base vocabulary (without the added tokens).
        """
        return self._tokenizer.get_vocab_size(with_added_tokens=False)

    @property
    def special_ids(self):
        return [self.token_to_id(t) for t in self.special_tokens]

    @property
    def pad_token(self):
        return self.config.pad_token

    @property
    def bos_token(self):
        return self.config.bos_token

    @property
    def eos_token(self):
        return self.config.eos_token

    @property
    def unk_token(self):
        return self.config.unk_token

    @property
    def mask_token(self):
        return self.config.mask_token

    @property
    def cls_token(self):
        return self.config.cls_token

    @property
    def sep_token(self):
        return self.config.sep_token

    @property
    def pad_token_id(self):
        return self.token_to_id(self.config.pad_token)

    @property
    def bos_token_id(self):
        return self.token_to_id(self.config.bos_token)

    @property
    def eos_token_id(self):
        return self.token_to_id(self.config.eos_token)

    @property
    def unk_token_id(self):
        return self.token_to_id(self.config.unk_token)

    @property
    def mask_token_id(self):
        return self.token_to_id(self.config.mask_token)

    @property
    def cls_token_id(self):
        return self.token_to_id(self.config.cls_token)

    @property
    def sep_token_id(self):
        return self.token_to_id(self.config.sep_token)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, List

import numpy as np

from ....preprocessors.image_processor import ImageProcessor, ImageProcessorConfig
from ....registry import register_preprocessor
from ....utils import (
    convert_batch_dict_dtype,
    convert_image_type,
    load_image,
    mirror_image,
    normalize_image,
    resize_image,
    transpose_channels_axis_side,
)


@dataclass
class CraftImageProcessorConfig(ImageProcessorConfig):
    name = "craft_image_processor"
    mean: tuple[float] = (123.675, 116.28, 103.53)
    std: tuple[float] = (58.395, 57.12, 57.375)
    rescale: float = None
    square_size: int = 2560
    mag_ratio: float = 1.0


@register_preprocessor("craft_image_processor", config_class=CraftImageProcessorConfig)
class CraftImageProcessor(ImageProcessor):
    def __init__(self, config: CraftImageProcessorConfig, **kwargs):
        super().__init__(config=config, **kwargs)

    def get_ratio(self, image, square_size: int = None, mag_ratio: float = None):
        square_size = square_size or self.config.square_size
        mag_ratio = mag_ratio or self.config.mag_ratio

        height, width, channel = image.shape

        target_size = mag_ratio * max(height, width)
        target_size = min(target_size, square_size)

        ratio = target_size / max(height, width)

        return ratio

    def _resize(self, image, square_size: int, mag_ratio: float):
        """
        Resizes the image by pasting it into a bigger canvas
        """
        height, width, channel = image.shape

        # magnify image size
        target_size = mag_ratio * max(height, width)

        # set original image size
        target_size = min(target_size, square_size)

        ratio = target_size / max(height, width)

        target_h, target_w = int(height * ratio), int(width * ratio)
        inner_image = resize_image(image, size=(target_w, target_h))

        # make canvas and paste image
        target_h32, target_w32 = target_h, target_w
        if target_h % 32 != 0:
            target_h32 = target_h + (32 - target_h % 32)
        if target_w % 32 != 0:
            target_w32 = target_w + (32 - target_w % 32)
        resized_image = np.zeros((target_h32, target_w32, channel), dtype=np.float32)
        resized_image[0:target_h, 0:target_w, :] = inner_image

        return resized_image

    def __call__(
        self,
        images: List,
        device: str = None,
        mean: float = None,
        std: float = None,
        rescale: float = None,
        square_size: tuple[int] = None,
        mag_ratio: float = None,
        mirror: bool = None,
        gray_scale: bool = None,
        return_tensors: str = "torch",
        **kwargs,
    ):
        mean = mean or self.config.mean
        std = std or self.config.std
        square_size = square_size or self.config.square_size
        mag_ratio = mag_ratio or self.config.mag_ratio
        mirror = mirror or self.config.mirror

        if not isinstance(images, Iterable) or isinstance(images, str):
            images = [images]

        # Load images if inputs are list of files
        images = [load_image(x, return_type="numpy") if isinstance(x, str) else x for x in images]

        # Cast image types
        images = [convert_image_type(image, target_type="numpy") for image in images]

        # Mirror images if mirror is set
        if mirror:
            images = [mirror_image(image, return_type="numpy") for image in images]

        ratio_values = [self.get_ratio(image) for image in images]
        images = [self._resize(image, square_size=square_size, mag_ratio=mag_ratio) for image in images]

        if mean is not None and std is not None:
            images = [normalize_image(image, mean=mean, std=std, channel_axis="last") for image in images]

        # Transpose channels axis
        images = [transpose_channels_axis_side(image, axis_side="first") for image in images]

        # Return images batch dict
        images = np.array([convert_image_type(image, target_type="numpy") for image in images], dtype=np.float32)

        outputs = convert_batch_dict_dtype({"pixel_values": images}, dtype=return_tensors)

        if device:
            import torch

            outputs = {k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in outputs.items()}

        outputs["ratio_values"] = ratio_values

        return outputs



from .craft_text_detection import CraftTextDetection
from .craft_text_detection_config import CraftTextDetectionConfig
from .craft_image_processor import CraftImageProcessor, CraftImageProcessorConfig



from dataclasses import dataclass

import torch

from ...configs import DatasetConfig
from ...constants import Backends, TaskType
from ...registry import register_dataset
from ...utils import Logger, is_backend_available, shift_tokens_right
from ..data_collators import ImageCaptioningDataCollator
from .dataset import Dataset


if is_backend_available(Backends.DATASETS):
    from datasets import load_dataset

logger = Logger(__name__)

_required_backends = [Backends.DATASETS]


@dataclass
class ImageCaptioningDatasetConfig(DatasetConfig):
    """
    Configuration class for image captioning datasets.

    Args:
        path (str): Path to the dataset.
        text_column (str): Column name for text in the dataset.
        images_paths_column (str): Column name for image paths in the dataset.
        max_length (int): Maximum length of text.

    """
    name = "image_captioning"
    task: TaskType = TaskType.IMAGE2TEXT
    path: str = None
    text_column: str = "label"
    images_paths_column = "image_path"
    max_length: int = None


@register_dataset("image_captioning", config_class=ImageCaptioningDatasetConfig)
class ImageCaptioningDataset(Dataset):
    required_backends = _required_backends

    def __init__(self, config: ImageCaptioningDatasetConfig, split=None, preprocessor=None, **kwargs):
        super().__init__(config=config, split=split, preprocessor=preprocessor, **kwargs)
        self.image_processor = self.preprocessor.image_processor
        self.tokenizer = self.preprocessor.tokenizer
        self.data_collator = ImageCaptioningDataCollator(
            self.tokenizer,
            padding="max_length" if self.config.max_length is not None else "longest",
            max_length=self.config.max_length
        )

    def _load(self, split):
        """
        Load the dataset and clean up invalid samples.

        Args:
            split: Dataset split, defaults to None.

        Returns:
            Dataset: The cleaned dataset.

        """
        data = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir, **self.config.hf_load_kwargs)
        return data

    def __getitem__(self, index):
        """
        Get a specific item from the dataset.

        Args:
            index: Index of the item to retrieve.

        Returns:
            dict: The input data.
        """
        path, text = self.data[index].values()
        pixel_values = self.image_processor(path, return_tensors="torch")["pixel_values"][0]
        tokenized_inputs = self.tokenizer(text, padding="max_length", max_length=self.config.max_length)
        decoder_input_ids = shift_tokens_right(
            [tokenized_inputs["token_ids"]],
            pad_token_id=self.tokenizer.pad_token_id,
            decoder_start_token_id=self.tokenizer.bos_token_id,
        )[0]
        inputs = {
            "pixel_values": pixel_values,
            "labels": tokenized_inputs["token_ids"],
            "decoder_input_ids": decoder_input_ids,
            "decoder_attention_mask": tokenized_inputs["attention_mask"],
        }
        return inputs



from dataclasses import dataclass

import torch

from ...configs import DatasetConfig
from ...constants import Backends, TaskType
from ...registry import register_dataset
from ...utils import Logger, is_backend_available
from ..data_collators import TextPaddingDataCollator
from .dataset import Dataset


if is_backend_available(Backends.DATASETS):
    from datasets import load_dataset

logger = Logger(__name__)


@dataclass
class TextClassificationDatasetConfig(DatasetConfig):
    """
    Configuration class for text classification datasets.

    Args:
        path (str): Path to the dataset.
        label_field (str): Field name for labels in the dataset.
        text_field (str): Field name for text in the dataset.
        max_length (int): Maximum length of text.
    """

    name = "text_classification"
    task: TaskType = TaskType.TEXT_CLASSIFICATION
    path: str = None
    label_field: str = None
    text_field: str = None
    max_length: int = None


@register_dataset("text_classification", config_class=TextClassificationDatasetConfig)
class TextClassificationDataset(Dataset):
    """
    A text classification dataset class.
    As of now this class is intended for datasets existing on the Hub!

    Args:
        config (TextClassificationDatasetConfig): Dataset config object.
        split: Which split to use.
        preprocessor: Dataset's preprocessor
        **kwargs: Extra config parameters to assign to the original config.
    """

    def __init__(self, config: TextClassificationDatasetConfig, split=None, preprocessor=None, **kwargs):
        super().__init__(config, split=split, preprocessor=preprocessor, **kwargs)
        self._extract_labels()
        self.tokenizer = self.preprocessor.tokenizer
        self.data_collator = TextPaddingDataCollator(
            tokenizer=self.tokenizer,
            max_length=self.config.max_length,
        ) if self.tokenizer else None

    def _load(self, split):
        """
        Load the dataset.

        Args:
            split: Dataset split.

        Returns:
            The whole dataset.

        """
        dataset = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir, **self.config.hf_load_kwargs)
        return dataset

    def _extract_labels(self):
        """
        Extract label names, ids and build dictionaries.
        """
        labels_list = self.data.features[self.config.label_field].names
        self.id2label = self.config.id2label = {k: str(v) for k, v in dict(enumerate(labels_list)).items()}
        self.label2id = self.config.label2id = {v: k for k, v in self.id2label.items()}
        self.num_labels = self.config.num_labels = len(labels_list)

    def __getitem__(self, index):
        """
        Tokenize inputs and return a dict containing ids, masks, labels, etc.

        Args:
            index: Sample index.

        Returns:
            dict: The input data.

        """
        text = self.data[index][self.config.text_field]
        label = self.data[index][self.config.label_field]
        inputs = self.tokenizer(text, return_attention_mask=True)
        inputs["labels"] = label

        return inputs



from __future__ import annotations

import math
import os
from typing import List

from torch.utils.data import Dataset as TorchDataset

from ...builders import build_dataset
from ...configs import DatasetConfig
from ...constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    HEZAR_CACHE_DIR,
    Backends,
    RepoType,
    SplitType,
)
from ...preprocessors import Preprocessor, PreprocessorsContainer
from ...utils import Logger, get_module_config_class, list_repo_files, verify_dependencies


logger = Logger(__name__)


class Dataset(TorchDataset):
    """
    Base class for all datasets in Hezar.

    Args:
        config: The configuration object for the dataset.
        split: Dataset split name e.g, train, test, validation, etc.
        preprocessor: Preprocessor object or path (note that Hezar datasets classes require this argument).
        **kwargs: Additional keyword arguments.

    Attributes:
        required_backends (List[str | Backends]): List of required backends for the dataset.
        config_filename (str): Default dataset config file name.
        cache_dir (str): Default cache directory for the dataset.

    """
    required_backends: List[str | Backends] = [Backends.DATASETS]
    config_filename = DEFAULT_DATASET_CONFIG_FILE
    cache_dir = os.path.join(HEZAR_CACHE_DIR, "datasets")

    def __init__(
        self,
        config: DatasetConfig,
        split: str = "train",
        preprocessor: str | Preprocessor | PreprocessorsContainer = None,
        **kwargs,
    ):
        verify_dependencies(self, self.required_backends)
        self.config = config.update(kwargs)
        self.split = split
        self.data = self._load(self.split)
        self.preprocessor = self.create_preprocessor(preprocessor)
        self.data_collator = None

    def _load(self, split):
        """
        The internal function to load the dataset files and properties.
        By default, this uses the HF `datasets.load_dataset()`.
        """
        pass

    @staticmethod
    def create_preprocessor(preprocessor: str | Preprocessor | PreprocessorsContainer):
        """
        Create the preprocessor for the dataset.

        Args:
            preprocessor (str | Preprocessor | PreprocessorsContainer): Preprocessor for the dataset
        """
        if preprocessor is None:
            logger.warning(
                "Since v0.39.0, `Dataset` classes require the `preprocessor` parameter and cannot be None or it will "
                "lead to errors later on! (This warning will change to an error in the future)"
            )
            return PreprocessorsContainer()

        if isinstance(preprocessor, str):
            preprocessor = Preprocessor.load(preprocessor)
        if isinstance(preprocessor, Preprocessor):
            preprocessor = PreprocessorsContainer({preprocessor.config.name: preprocessor})
        elif isinstance(preprocessor, PreprocessorsContainer):
            preprocessor = preprocessor
        else:
            raise ValueError(
                f"The `preprocessor` must be a path to the Hub or a Preprocessor/PreprocessorsContainer instance, "
                f"got {type(preprocessor)}!"
            )
        return preprocessor

    def __str__(self):
        dataset_name = self.config.path or self.config.name
        dataset_size = len(self)
        return f"{self.__class__.__name__}(path={dataset_name}['{self.split}'], size={dataset_size})"

    def __len__(self):
        """
        Returns the length of the dataset. The `max_size` parameter in the config can overwrite this value. Override
        with caution!
        """
        if isinstance(self.config.max_size, float) and 0 < self.config.max_size <= 1:
            return math.ceil(self.config.max_size * len(self.data))

        elif isinstance(self.config.max_size, int) and 0 < self.config.max_size < len(self.data):
            return self.config.max_size

        return len(self.data)

    def __getitem__(self, index):
        """
        Gets a specific item from the dataset.

        Args:
            index: Index of the item to retrieve.

        Raises:
            NotImplementedError: This method must be implemented in derived classes.

        """
        raise NotImplementedError

    @classmethod
    def load(
        cls,
        hub_path: str | os.PathLike,
        split: str | SplitType = None,
        preprocessor: str | Preprocessor | PreprocessorsContainer = None,
        config: DatasetConfig = None,
        config_filename: str = None,
        cache_dir: str = None,
        **kwargs,
    ) -> "Dataset":
        """
        Load the dataset from a hub path.

        Args:
            hub_path (str | os.PathLike):
                Path to dataset from hub or locally.
            split (Optional[str | SplitType]):
                Dataset split, defaults to "train".
            preprocessor (str | Preprocessor | PreprocessorsContainer):
                Preprocessor object for the dataset
            config: (DatasetConfig):
                A config object to ignore the config in the repo or in case the repo has no `dataset_config.yaml` file
            config_filename (Optional[str]):
                Dataset config file name. Falls back to `dataset_config.yaml` if not given.
            cache_dir (str):
                Path to cache directory, defaults to Hezar's cache directory
            **kwargs:
                Config parameters as keyword arguments.

        Returns:
            Dataset: An instance of the loaded dataset.

        """
        split = split or "train"
        config_filename = config_filename or cls.config_filename

        if ":" in hub_path:
            hub_path, hf_dataset_config_name = hub_path.split(":")
            kwargs["hf_load_kwargs"] = kwargs.get("hf_load_kwargs", {})
            kwargs["hf_load_kwargs"]["name"] = hf_dataset_config_name

        if cache_dir is not None:
            cls.cache_dir = cache_dir

        has_config = config_filename in list_repo_files(hub_path, repo_type="dataset")

        if config is not None:
            dataset_config = config.update(kwargs)
        elif has_config:
            dataset_config = DatasetConfig.load(
                hub_path,
                filename=config_filename,
                repo_type=RepoType.DATASET,
                cache_dir=cls.cache_dir,
                **kwargs,
            )
        elif kwargs.get("task", None):
            config_cls = get_module_config_class(kwargs["task"], registry_type="dataset")
            if config_cls:
                dataset_config = config_cls(**kwargs)
            else:
                raise ValueError(f"Task `{kwargs['task']}` is not valid!")
        else:
            raise ValueError(
                f"The dataset at `{hub_path}` does not have enough info and config to load using Hezar!"
                f"\nHint: Either pass the proper `config` to `.load()` or pass in required config parameters as "
                f"kwargs in `.load()`, most notably `task`!"
            )

        dataset_config.path = hub_path
        dataset = build_dataset(
            dataset_config.name,
            config=dataset_config,
            split=split,
            preprocessor=preprocessor,
            **kwargs,
        )
        return dataset



import random
from typing import Sized

from torch.utils.data import Sampler

from ..utils import set_seed


__all__ = [
    "RangedSampler",
]


class RangedSampler(Sampler):
    """
    A simple dataset sampler that samples the data from a starting index up to an ending index and also supports
    shuffling, drop_last, etc.

    Args:
        data: The dataset source
        batch_size: Batch size
        start_index: The starting index, will default to 0 if not given
        end_index: The ending index, will default to the length of the data source
        drop_last: Whether to drop the indices of the last batch
        shuffle: Whether to shuffle indices
        seed: Seed value for shuffling. Required if shuffle is True
    """

    def __init__(
        self,
        data: Sized,
        batch_size: int,
        start_index: int = None,
        end_index: int = None,
        drop_last: bool = False,
        shuffle: bool = False,
        seed: int = None
    ):
        super().__init__()
        self.data = data
        self.start_index = start_index or 0
        self.end_index = end_index or len(self.data)
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.shuffle = shuffle
        self.seed = seed

        self.end_index = (self.end_index // batch_size) * batch_size if self.drop_last else self.end_index
        self.num_samples = self.end_index - self.start_index
        self.total_length = len(self.data)

        self.indices = self._create_indices()

    def _create_indices(self):
        indices = list(range(self.total_length))
        if self.shuffle:
            if self.seed is None:
                raise ValueError("Parameter `seed` is required to enable shuffling in the data sampler!")
            set_seed(self.seed)
            random.shuffle(indices)
        indices = indices[self.start_index:self.end_index]
        return indices

    def __len__(self):
        return self.total_length

    def __iter__(self):
        for indice in self.indices:
            yield indice



"""
Dataset processors are a bunch of callable classes to be passed as map functions for any dataset on the Hub.
Note that the main dataset classes are already implemented in a way that the processing is done in the `__getitem__`
method and these classes are only used for when the dataset has been loaded using the HuggingFace datasets library,
and you want to get advantage of the multiprocessing/batch processing/caching functionalities of the HF datasets.

Example:
>>> from datasets import load_dataset
>>> from hezar.data import SpeechRecognitionDatasetProcessor

>>> data_processor = SpeechRecognitionDatasetProcessor(feature_extractor=feature_extractor,tokenizer=tokenizer)
>>> dataset = load_dataset("hezarai/common-voice-13-fa")
>>> dataset = dataset.map(data_processor, batched=True, batch_size=1000)
"""

import torch

from ..constants import Backends
from ..utils import is_backend_available, reverse_string_digits, verify_dependencies


if is_backend_available(Backends.DATASETS):
    from datasets.formatting.formatting import LazyBatch, LazyRow

__all__ = [
    "DatasetProcessor",
    "ImageCaptioningDatasetProcessor",
    "OCRDatasetProcessor",
    "SequenceLabelingDatasetProcessor",
    "SpeechRecognitionDatasetProcessor",
    "TextClassificationDatasetProcessor",
    "TextSummarizationDatasetProcessor",
]


class DatasetProcessor:
    """
    The base callable dataset processor class that can handle both single and batched mode dataset mapping.
    """
    required_backends = [Backends.DATASETS]

    def __init__(self, *args, **kwargs):
        verify_dependencies(self, self.required_backends)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, data: LazyBatch | LazyRow, return_tensors="list", **kwargs):
        """
        Method called when using the map function.
        Decides whether to call `process_single()` or `process_batch()` based on the data values.

        Args:
            data: A dict of feature name -> sample or batch of samples mapping.
            return_tensors: The type of the returning tensors (list, torch, numpy)
            **kwargs: Additional keyword arguments passed through the `map` function as `kwargs`.
        """
        if isinstance(data, LazyRow):
            return self.process_single(data, return_tensors=return_tensors, **kwargs)
        elif isinstance(data, LazyBatch):
            return self.process_batch(data, return_tensors=return_tensors, **kwargs)
        else:
            raise ValueError(f"The input data must be either `LazyBatch` or `LazyRow`, got `{type(data)}`!")

    def process_single(self, data: LazyRow, return_tensors=None, **kwargs):
        """
        Process a single data example.

        Args:
            data: A data sample dict
            return_tensors: The type of the returning tensors (list, torch, numpy)
            **kwargs: Additional arguments

        Returns:
            The updated data dict
        """
        raise NotImplementedError

    def process_batch(self, data: LazyBatch, return_tensors=None, **kwargs):
        """
        Process a batch of data examples.

        Args:
            data: A data sample dict
            return_tensors: The type of the returning tensors (list, torch, numpy)
            **kwargs: Additional arguments

        Returns:
            The updated data dict
        """
        raise NotImplementedError


class ImageCaptioningDatasetProcessor(DatasetProcessor):
    """
    Dataset processor for image captioning datasets. This class handles tokenization and image processing.
    """

    def __init__(self, image_processor, tokenizer, max_length=None, padding=None):
        super().__init__()
        self.image_processor = image_processor
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.padding = padding

    @staticmethod
    def _shift_tokens_right(input_ids: list[list[int]], pad_token_id: int, decoder_start_token_id: int):
        """
        Shift input ids one token to the right.
        """
        # Initialize shifted_input_ids as a list of lists with the same shape as input_ids
        shifted_input_ids = [[0] * len(row) for row in input_ids]

        for i, row in enumerate(input_ids):
            # Shift each row one token to the right
            shifted_input_ids[i][1:] = row[:-1]
            # Set the first token to decoder_start_token_id
            shifted_input_ids[i][0] = decoder_start_token_id

            # Replace any -100 values with pad_token_id
            shifted_input_ids[i] = [pad_token_id if token == -100 else token for token in shifted_input_ids[i]]

        return shifted_input_ids

    def process_single(self, data, return_tensors=None, padding=None, max_length=None):
        """
        Process image and tokenize captions for a single data sample.

        Args:
            data: A data example containing the image and its caption
            padding: Padding type e.g, max_length, longest.
            max_length: Max length value if padding is set to max_length or the labels must be truncated.
            return_tensors: The type of the returning tensors (list, torch, numpy)

        Returns:
            A dict of pixel values tensor of the processed image and labels token ids and attention mask.
        """
        padding = padding or self.padding
        max_length = max_length or self.max_length

        path = data["image_path"]
        text = data["label"]

        tokenized_inputs = self.tokenizer(text, padding=padding, max_length=max_length, return_tensors=return_tensors)

        data["pixel_values"] = self.image_processor(path, return_tensors=return_tensors)["pixel_values"]
        data["labels"] = tokenized_inputs["token_ids"]
        data["decoder_attention_mask"] = tokenized_inputs["attention_mask"]
        data["decoder_input_ids"] = self._shift_tokens_right(
            [data["labels"]],
            pad_token_id=self.tokenizer.pad_token_id,
            decoder_start_token_id=self.tokenizer.bos_token_id,
        )[0]

        return data

    def process_batch(self, data, return_tensors=None, padding=None, max_length=None):
        """
        Process image and tokenize captions for a batch of data samples.

        Args:
            data: A batch of data examples containing the images and their captions
            padding: Padding type e.g, max_length, longest.
            max_length: Max length value if padding is set to max_length or the labels must be truncated.
            return_tensors: The type of the returning tensors (list, torch, numpy)

        Returns:
            A dict of pixel values tensor of the processed images and labels token ids and attention masks.
        """
        padding = padding or self.padding
        max_length = max_length or self.max_length

        paths = data["image_path"]
        texts = data["label"]

        tokenized_inputs = self.tokenizer(texts, padding=padding, max_length=max_length, return_tensors=return_tensors)

        data["pixel_values"] = self.image_processor(paths, return_tensors=return_tensors)["pixel_values"]
        data["labels"] = tokenized_inputs["token_ids"]
        data["decoder_attention_mask"] = tokenized_inputs["attention_mask"]
        data["decoder_input_ids"] = self._shift_tokens_right(
            data["labels"],
            pad_token_id=self.tokenizer.pad_token_id,
            decoder_start_token_id=self.tokenizer.bos_token_id,
        )

        return data


class OCRDatasetProcessor(DatasetProcessor):
    """
    Dataset processor class for OCR which can handle both tokenizer-based or character-split-based datasets.
    """

    def __init__(
        self,
        image_processor,
        tokenizer=None,
        text_split_type="char_split",
        max_length=None,
        reverse_digits=False,
        id2label=None,
        image_field="image_path",
        text_field="text",
    ):
        super().__init__()
        self.image_processor = image_processor
        self.tokenizer = tokenizer
        self.text_split_type = text_split_type
        self.max_length = max_length
        self.reverse_digits = reverse_digits
        self.id2label = id2label
        self.image_field = image_field
        self.text_field = text_field

    def _text_to_ids(self, text):
        """
        Convert text to tensor based on the configured text_split_type.

        Args:
            text (str): The raw text.

        Returns:
            torch.Tensor: The output tensor.

        """
        if self.text_split_type == "tokenize":
            token_ids = self.tokenizer(text, padding="max_length", max_length=self.max_length)["input_ids"]
            labels = [token_id if token_id != self.tokenizer.pad_token_id else -100 for token_id in token_ids]
        elif self.text_split_type == "char_split":
            if self.reverse_digits:
                text = reverse_string_digits(text)
            label2id = {v: k for k, v in self.id2label.items()}
            labels = [label2id[char] for char in text]
        else:
            raise ValueError(f"Invalid `text_split_type={self.text_split_type}`")
        return labels

    def process_single(self, data, return_tensors=None):
        """
        Process a single image-to-text OCR example.

        Args:
            data: A data example containing an image path and corresponding text.
            return_tensors: The type of the returning tensors (list, torch, numpy)

        Returns:
            dict: Processed inputs with pixel values and text labels.
        """
        path = data[self.image_field]
        text = data[self.text_field]
        pixel_values = self.image_processor(path, return_tensors=return_tensors)["pixel_values"][0]
        labels = self._text_to_ids(text)
        return {"pixel_values": pixel_values, "labels": labels}

    def process_batch(self, data, return_tensors=None):
        """
        Process a batch of image-to-text OCR examples.

        Args:
            data: A batch of data examples containing image paths and corresponding texts.
            return_tensors: The type of the returning tensors (list, torch, numpy)

        Returns:
            dict: Batch of processed inputs with pixel values and text labels.
        """
        paths = data[self.image_field]
        texts = data[self.text_field]

        # Process images in batch
        pixel_values = self.image_processor(paths, return_tensors=return_tensors)["pixel_values"]

        # Process text labels in batch
        labels = [self._text_to_ids(text) for text in texts]

        return {"pixel_values": pixel_values, "labels": labels}


class SequenceLabelingDatasetProcessor(DatasetProcessor):
    """
    Dataset processor class for sequence labeling datasets. Handles tokenization and label alignment.
    """

    def __init__(self, tokenizer, label_all_tokens=True, ignore_index=-100, max_length=None, padding=None):
        super().__init__()
        self.tokenizer = tokenizer
        self.label_all_tokens = label_all_tokens
        self.ignore_index = ignore_index
        self.max_length = max_length
        self.padding = padding

    def _tokenize_and_align(self, tokens, labels, return_tensors=None, padding=None, max_length=None):
        """
        Tokenize and align tokens and labels for sequence labeling tasks.

        Args:
            tokens: List of tokens (for single examples) or list of lists (for batches).
            labels: List of labels (for single examples) or list of lists (for batches).
            return_tensors: The type of the returning tensors (list, torch, numpy)
            padding: Padding strategy for tokenization.
            max_length: Maximum sequence length to truncate/pad.

        Returns:
            dict: Tokenized and aligned inputs with labels.
        """
        padding = padding or self.padding
        max_length = max_length or self.max_length

        # Tokenize and return word IDs for mapping labels to subword tokens
        tokenized_inputs = self.tokenizer(
            tokens,
            is_split_into_words=True,
            return_word_ids=True,
            padding=padding,
            truncation=True,
            max_length=max_length,
            return_tensors=return_tensors
        )
        word_ids = tokenized_inputs["word_ids"]

        # Align labels with tokens
        aligned_labels = []
        for batch_idx, batch_word_ids in enumerate(word_ids):
            previous_word_idx = None
            label_ids = []
            for word_idx in batch_word_ids:
                # Assign ignore index for special tokens
                if word_idx is None:
                    label_ids.append(self.ignore_index)
                elif word_idx != previous_word_idx:
                    # Assign the label for the first token of each word
                    label_ids.append(labels[batch_idx][word_idx])
                else:
                    # Assign label for subword tokens (if label_all_tokens is True)
                    label_ids.append(labels[batch_idx][word_idx] if self.label_all_tokens else self.ignore_index)
                previous_word_idx = word_idx
            aligned_labels.append(label_ids)

        tokenized_inputs["labels"] = aligned_labels
        return tokenized_inputs

    def process_single(self, data, return_tensors=None, padding=None, max_length=None):
        """
        Process a single example of sequence labeling data.

        Args:
            data: A single data example containing tokens and labels.
            return_tensors: The type of the returning tensors (list, torch, numpy)
            padding: Padding strategy.
            max_length: Maximum sequence length.

        Returns:
            dict: Tokenized and aligned input data.
        """
        tokens = data["tokens"]
        labels = data["pos_tags"]

        tokenized_inputs = self._tokenize_and_align(
            [tokens],
            [labels],
            return_tensors=return_tensors,
            padding=padding,
            max_length=max_length,
        )
        tokenized_inputs = {k: v[0] for k, v in tokenized_inputs.items()}
        data.update(tokenized_inputs)

        return data

    def process_batch(self, data, return_tensors=None, padding=None, max_length=None):
        """
        Process a batch of sequence labeling examples.

        Args:
            data: A batch of examples, containing tokens and labels.
            return_tensors: The type of the returning tensors (list, torch, numpy)
            padding: Padding strategy.
            max_length: Maximum sequence length.

        Returns:
            dict: Tokenized and aligned batch data.
        """
        tokens = data["tokens"]
        labels = data["pos_tags"]

        tokenized_inputs = self._tokenize_and_align(
            tokens,
            labels,
            return_tensors=return_tensors,
            padding=padding,
            max_length=max_length,
        )

        data.update(tokenized_inputs)

        return data


class SpeechRecognitionDatasetProcessor(DatasetProcessor):
    """
    Processor class for speech recognition datasets. Handles audio feature extraction and labels tokenization.
    """

    def __init__(
        self,
        feature_extractor,
        tokenizer,
        sampling_rate=16000,
        audio_array_padding=None,
        max_audio_array_length=None,
        labels_padding=None,
        labels_max_length=None,
        audio_column="audio",
        transcript_column="transcript",
    ):
        super().__init__()
        self.feature_extractor = feature_extractor
        self.tokenizer = tokenizer
        self.sampling_rate = sampling_rate
        self.audio_array_padding = audio_array_padding
        self.max_audio_array_length = max_audio_array_length
        self.labels_padding = labels_padding
        self.labels_max_length = labels_max_length
        self.audio_column = audio_column
        self.transcript_column = transcript_column

    def process_single(self, data, return_tensors=None):
        """
        Process a single speech recognition example.

        Args:
            data: A data example containing audio and its transcript.
            return_tensors: The type of the returning tensors (list, torch, numpy)

        Returns:
            dict: Processed input features and labels.
        """
        audio_array = data[self.audio_column]["array"]
        transcript = data[self.transcript_column]

        # Extract input features from audio
        input_features = self.feature_extractor(
            audio_array,
            sampling_rate=self.sampling_rate,
            padding=self.audio_array_padding,
            max_length=self.max_audio_array_length,
            return_tensors=return_tensors,
        )["input_features"]

        # Tokenize the transcript
        labels = self.tokenizer(
            transcript,
            padding=self.labels_padding,
            max_length=self.labels_max_length,
            return_tensors=return_tensors,
        )

        data["input_features"] = input_features
        data["labels"] = labels["token_ids"]
        data["attention_mask"] = labels["attention_mask"]

        return data

    def process_batch(self, data, return_tensors=None):
        """
        Process a batch of speech recognition examples.

        Args:
            data: A batch of data examples containing audio arrays and their corresponding transcripts.
            return_tensors: The type of the returning tensors (list, torch, numpy)

        Returns:
            dict: Batch of processed input features and labels.
        """
        audio_arrays = [x["array"] for x in data[self.audio_column]]
        transcripts = data[self.transcript_column]

        # Extract input features in batch
        input_features = self.feature_extractor(
            audio_arrays,
            sampling_rate=self.sampling_rate,
            padding=self.audio_array_padding,
            max_length=self.max_audio_array_length,
            return_tensors=return_tensors,
        )["input_features"]

        # Tokenize transcripts in batch
        labels = self.tokenizer(
            transcripts,
            padding=self.labels_padding,
            max_length=self.labels_max_length,
            return_tensors=return_tensors,
        )

        data["input_features"] = input_features
        data["labels"] = labels["token_ids"]
        data["attention_mask"] = labels["attention_mask"]

        return data


class TextClassificationDatasetProcessor(DatasetProcessor):
    """
    Processor class for text classification datasets. Handles tokenization of the texts.
    """

    def __init__(self, tokenizer, max_length=None, padding=None):
        super().__init__()
        self.tokenizer = tokenizer
        self.padding = padding
        self.max_length = max_length

    def process_single(self, data, return_tensors=None, padding=None, max_length=None):
        """
        Process a single example for text classification.

        Args:
            data: A single data example dict
            return_tensors: The type of the returning tensors (list, torch, numpy)
            padding: Token ids padding type
            max_length: Max input length

        Returns:
            The updated data dictionary
        """
        padding = padding or self.padding
        max_length = max_length or self.max_length

        text = data["text"]
        label = data["label"]

        inputs = self.tokenizer(
            text,
            padding=padding,
            max_length=max_length,
            return_attention_mask=True,
            return_tensors=return_tensors,
        )
        data.update(inputs)
        data["labels"] = torch.tensor(label, dtype=torch.long)

        return data

    def process_batch(self, data, return_tensors=None, padding=None, max_length=None):
        """
        Process a batch of examples for text classification.

        Args:
            data: A single data example dict
            return_tensors: The type of the returning tensors (list, torch, numpy)
            padding: Token ids padding type
            max_length: Max input length

        Returns:
            The updated data dictionary
        """
        padding = padding or self.padding
        max_length = max_length or self.max_length

        texts = data["text"]
        labels = data["label"]

        inputs = self.tokenizer(
            texts,
            padding=padding,
            max_length=max_length,
            return_attention_mask=True,
            return_tensors=return_tensors,
        )
        data.update(inputs)
        data["labels"] = torch.tensor(labels, dtype=torch.long)

        return data


class TextSummarizationDatasetProcessor(DatasetProcessor):
    """
    Processor class for text summarization datasets. Handles tokenization of the inputs and labels.
    """

    def __init__(
        self,
        tokenizer,
        prefix=None,
        max_length=None,
        labels_max_length=None,
        text_field="text",
        summary_field="summary",
        padding=None,
    ):
        super().__init__()
        self.tokenizer = tokenizer
        self.prefix = prefix
        self.max_length = max_length
        self.labels_max_length = labels_max_length
        self.text_field = text_field
        self.summary_field = summary_field
        self.padding = padding

    def process_single(self, data, return_tensors=None, padding=None, max_length=None, labels_max_length=None):
        """
        Process a single example for text summarization.

        Args:
            data: A data example containing text and summary.
            return_tensors: The type of the returning tensors (list, torch, numpy)
            padding: Padding strategy.
            max_length: Max length for input text.
            labels_max_length: Max length for summary labels.

        Returns:
            dict: Tokenized inputs and labels for summarization task.
        """
        padding = padding or self.padding
        max_length = max_length or self.max_length
        labels_max_length = labels_max_length or self.labels_max_length

        text = data[self.text_field]
        summary = data[self.summary_field]

        # Add prefix if needed for conditional generation
        if self.prefix is not None:
            text = self.prefix + text

        # Tokenize inputs and labels
        inputs = self.tokenizer(
            text,
            max_length=max_length,
            padding=padding,
            return_attention_mask=True,
            return_tensors=return_tensors,
        )
        labels = self.tokenizer(
            summary,
            max_length=labels_max_length,
            padding=padding,
            return_attention_mask=True,
            return_tensors=return_tensors,
        )

        inputs["labels"] = labels["token_ids"]

        return inputs

    def process_batch(self, data, return_tensors=None, padding=None, max_length=None, labels_max_length=None):
        """
        Process a batch of examples for text summarization.

        Args:
            data: A batch of examples containing texts and summaries.
            return_tensors: The type of the returning tensors (list, torch, numpy)
            padding: Padding strategy.
            max_length: Max length for input texts.
            labels_max_length: Max length for summary labels.

        Returns:
            dict: Tokenized inputs and labels for summarization task.
        """
        padding = padding or self.padding
        max_length = max_length or self.max_length
        labels_max_length = labels_max_length or self.labels_max_length

        texts = data[self.text_field]
        summaries = data[self.summary_field]

        # Add prefix if needed for conditional generation
        if self.prefix is not None:
            texts = [self.prefix + text for text in texts]

        # Tokenize inputs and labels in batch
        inputs = self.tokenizer(
            texts,
            max_length=max_length,
            padding=padding,
            return_attention_mask=True,
            return_tensors=return_tensors,
        )
        labels = self.tokenizer(
            summaries,
            max_length=labels_max_length,
            padding=padding,
            return_attention_mask=True,
            return_tensors=return_tensors,
        )

        inputs["labels"] = labels["token_ids"]

        return inputs
